// fft.cpp: implementation of the Cxfft class.
//  This is a slightly modified version of Takuya OOURA's
//     original radix 4 FFT package.
//Copyright(C) 1996-1998 Takuya OOURA
//    (email: ooura@mmm.t.u-tokyo.ac.jp).
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>
#include "cfft.h"

BOOL testflag = FALSE;

//////////////////////////////////////////////////////////////////////
// Local Defines
//////////////////////////////////////////////////////////////////////
#define K_2PI (8.0*atan(1))


#define K_AMPMAX 2.14748e9	//maximum sin wave Pk
#define K_MAXDB 0.0
#define K_MINDB -160.0

#define NORMAL_MODE 0
#define MAX_MODE 1
#define DELTA_MODE 2
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
Cxfft::Cxfft()
{
	InitializeCriticalSection(&m_CriticalSection);
	m_Overload = FALSE;
	m_Invert = FALSE;
	m_MaxDelta = -1;
	m_NCOnull = FALSE;
	m_Complex = FALSE;
	m_AveSize = 1;
	m_LastAve = 1;
	m_LastFFTSize = 0;
	m_AveCount = 0;
	m_TotalCount = 0;
	m_PostAverage;
	m_LastWindow = -1;
	m_FFTSize = 1024;
	m_Gain = 1.0;
	pWorkArea = NULL;
	pSinCosTbl = NULL;
	pWindowTbl = NULL;
	pFFTAveBuf = NULL;
	pFFTInBuf = NULL;
	pFFTSumBuf = NULL;
	pFFTSecBuf = NULL;
	m_LastWindow = -1;
	m_MaxdB = K_MAXDB;
	SetFFTParams( FALSE, 1024, -140.0, 0.0 ,1 ,0, WINDOW_HANNING, FALSE, FALSE,NORMAL_MODE,0.0);
}

Cxfft::~Cxfft()
{							// free all resources
	DeleteCriticalSection(&m_CriticalSection);
	FreeMemory();
	if(pFFTSecBuf)
	{
		delete pFFTSecBuf;
		pFFTSecBuf = NULL;
	}
}

void Cxfft::FreeMemory()
{
	if(pWorkArea)
	{
		delete pWorkArea;
		pWorkArea = NULL;
	}
	if(pSinCosTbl)
	{
		delete pSinCosTbl;
		pSinCosTbl = NULL;
	}
	if(pWindowTbl)
	{
		delete pWindowTbl;
		pWindowTbl = NULL;
	}
	if(pFFTAveBuf)
	{
		delete pFFTAveBuf;
		pFFTAveBuf = NULL;
	}
	if(pFFTSumBuf)
	{
		delete pFFTSumBuf;
		pFFTSumBuf = NULL;
	}
	if(pFFTInBuf)
	{
		delete pFFTInBuf;
		pFFTInBuf = NULL;
	}
	if(pFFTSecBuf)
	{
		delete pFFTSecBuf;
		pFFTSecBuf = NULL;
	}
	
}

void Cxfft::SetFFTParams(BOOL Complex, INT size, double dBmin, double dBmax, INT ave,
						INT postave, INT Window, BOOL invert, BOOL NCOnull, INT MaxDelta,double maxdB)
{
INT i;
	EnterCriticalSection(&m_CriticalSection);
	m_Invert = invert;
	m_NCOnull = NCOnull;
	if(( m_Complex != Complex)	//if changed real or complex then force
		||(m_MaxDelta != MaxDelta) 
		|| (m_MaxdB != maxdB) )
	{							//reset of FFT params
		m_LastFFTSize = 0;
		m_LastAve = 0;
		m_MaxdB = maxdB;
	}
	m_Complex = Complex;
	m_MaxDelta = MaxDelta;
	if(ave>0)
		m_AveSize = ave;
	else
		m_AveSize = 1;
	m_PostAverage = (postave*size)/2000;
	m_PostAverage &= ~(0x00000001);
	LeaveCriticalSection(&m_CriticalSection);
	if( (size<=MAX_FFT_SIZE) && (size>=MIN_FFT_SIZE) )
		m_FFTSize = size;
	else
		m_FFTSize = MAX_FFT_SIZE;
	if(m_LastFFTSize != m_FFTSize)
	{
		EnterCriticalSection(&m_CriticalSection);
		m_LastFFTSize = m_FFTSize;
		m_LastWindow = -1;
		FreeMemory();
		pWindowTbl = new double[m_FFTSize];
		pSinCosTbl = new double[m_FFTSize/2];
		pWorkArea = new int[ (int)sqrt((double)m_FFTSize)+2];
		pFFTAveBuf = new double[m_FFTSize];
		pFFTSumBuf = new double[m_FFTSize];
		for(i=0; i<m_FFTSize; i++)
		{
			pFFTAveBuf[i] = 0.0;
			pFFTSumBuf[i] = 0.0;
		}
		if(m_MaxDelta != NORMAL_MODE)
		{
			pFFTSecBuf = new double[m_FFTSize];
			ResetPeaks();
		}
		pWorkArea[0] = 0;
		if(m_Complex)
		{
			pFFTInBuf = new double[m_FFTSize*2];
			for(i=0; i<m_FFTSize*2; i++)
				pFFTInBuf[i] = 0.0;
			makewt(m_FFTSize/2, pWorkArea, pSinCosTbl);
		}
		else
		{
			pFFTInBuf = new double[m_FFTSize*1];
			for(i=0; i<m_FFTSize; i++)
				pFFTInBuf[i] = 0.0;
			makewt(m_FFTSize/4, pWorkArea, pSinCosTbl);
			makect(m_FFTSize/4, pWorkArea, pSinCosTbl + pWorkArea[0]);
		}

//////////////////////////////////////////////////////////////////////
// A pure input sin wave ... Asin(wt)... will produce an fft output 
//   peak of (N*A/1)^2  where N is FFT_SIZE.
//		Kx = 2 for complex, 4 for real FFT
// To convert to a Power dB range:
//   PdBmax = 10*log10( (N*A/Kx)^2 + K_C ) + K_B
//   PdBmin = 10*log10( 0 + K_C ) + K_B
//  if (N*A/Kx)^2 >> K_C 
//  Then K_B = PdBmax - 20*log10( N*A/Kx )
//       K_C = 10 ^ ( (PdBmin-K_B)/10 )
//  for power range of 0 to 100 dB with input(A) of 32767 and N=262144
//			K_B = -86.63833  and K_C = 4.6114145e8
// To eliminate the multiply by 10, divide by 10 so for an output
//		range of 0 to -120dB the stored value is 0.0 to -12.0
//   so final constant K_B = -8.663833		
///////////////////////////////////////////////////////////////////////
		if(m_Complex)
			m_K_B = m_MaxdB - 20*log10( (double)m_FFTSize*K_AMPMAX/2.0 );
		else
			m_K_B = m_MaxdB - 20*log10( (double)m_FFTSize*K_AMPMAX/4.0 );
		m_K_C = pow( 10.0, (K_MINDB-m_K_B)/10.0 );
		m_K_B = m_K_B/10.0;
		LeaveCriticalSection(&m_CriticalSection);
		ResetFFT();
	}
	if( Window != m_LastWindow)
	{	// Pick a data windowing function:
		EnterCriticalSection(&m_CriticalSection);
		m_LastWindow = Window;
		switch(Window)
		{
			case WINDOW_RECTANGLE:
				for(i=0; i<m_FFTSize; i++)
					pWindowTbl[i] = .5;			//rectangle
				break;
			case WINDOW_HAMMING:
				for(i=0; i<m_FFTSize; i++)
					pWindowTbl[i] = (.54 - .46*cos( (K_2PI*i)/(m_FFTSize-1) ));	//Hamming
				break;
			case WINDOW_HANNING:
				for(i=0; i<m_FFTSize; i++)
					pWindowTbl[i] = (.5  - .5 *cos( (K_2PI*i)/(m_FFTSize-1) ));	//Hanning
				break;
		}
		LeaveCriticalSection(&m_CriticalSection);
		ResetFFT();
	}
	if(m_LastAve != m_AveSize)
	{
		m_LastAve = m_AveSize;
		ResetFFT();
	}
	EnterCriticalSection(&m_CriticalSection);
	m_dBDeltaOffset = dBmax/10.0;
	m_dBmaxOffset = dBmax/10.0;
	m_dBGainFactor = 10.0/(dBmax-dBmin);
	LeaveCriticalSection(&m_CriticalSection);
}

void Cxfft::ResetFFT()
{
	EnterCriticalSection(&m_CriticalSection);
	for(INT i=0; i<m_FFTSize;i++)
	{
		pFFTAveBuf[i] = 0.0;
		pFFTSumBuf[i] = 0.0;
	}
	m_SaveReference = FALSE;
	m_NeedToSaveReference = FALSE;
	m_AveCount = 0;
	m_TotalCount = 0;
	LeaveCriticalSection(&m_CriticalSection);
}

//////////////////////////////////////////////////////////////////////
// The function resets the peak display buffer
//////////////////////////////////////////////////////////////////////
void Cxfft::ResetPeaks()
{
	EnterCriticalSection(&m_CriticalSection);
	if(pFFTSecBuf)
	{
		for( int i=0; i<m_FFTSize; i++)
			pFFTSecBuf[i] = K_MINDB/10.;
	}
	LeaveCriticalSection(&m_CriticalSection);
}

void Cxfft::SaveReference()
{
	EnterCriticalSection(&m_CriticalSection);
	m_NeedToSaveReference = TRUE;
	LeaveCriticalSection(&m_CriticalSection);
}

//////////////////////////////////////////////////////////////////////
// "InBuf[]" is first multiplied by a window function and then
//  calculates an "FFT_SIZE" point FFT on "InBuf[]".
//////////////////////////////////////////////////////////////////////
INT Cxfft::SendFFT(_int32* InBuf, INT OVLimit)
{
INT i,j;
double dtmp;
	m_Overload = FALSE;
	EnterCriticalSection(&m_CriticalSection);
	if(m_NeedToSaveReference)
	{
		m_NeedToSaveReference = FALSE;
		m_SaveReference = TRUE;	//trigger save
	}
	if(m_Complex)
	{
		j=0;
		for(i=0; i<m_FFTSize; i++)
		{
			if( InBuf[j] > OVLimit )	//flag overload if within OVLimit of max
				m_Overload = TRUE;
			dtmp = pWindowTbl[i];
			pFFTInBuf[j] = dtmp * (double)InBuf[j];	//window the I data
			j++;
			pFFTInBuf[j] = dtmp * (double)InBuf[j];	//window the Q data
			j++;
		}
	}
	else
	{
		for(i=0; i<m_FFTSize; i++)
		{
			if( InBuf[i] > OVLimit )	//flag overload if within OVLimit of max
				m_Overload = TRUE;
			pFFTInBuf[i] = pWindowTbl[i] * (double)InBuf[i];		//window the data
		}
	}
	LeaveCriticalSection(&m_CriticalSection);
	return m_TotalCount;
}
//////////////////////////////////////////////////////////////////////
// "InBuf[]" is first multiplied by a window function and then
//  calculates an "FFT_SIZE" point FFT on "InBuf[]".
//////////////////////////////////////////////////////////////////////
INT Cxfft::SendFFT(SHORT* InBuf, INT OVLimit)
{
INT i,j;
double dtmp;
	m_Overload = FALSE;
	EnterCriticalSection(&m_CriticalSection);
	if(m_NeedToSaveReference)
	{
		m_NeedToSaveReference = FALSE;
		m_SaveReference = TRUE;	//trigger save
	}
	if(m_Complex)
	{
		j=0;
		for(i=0; i<m_FFTSize; i++)
		{
			if( InBuf[j] > OVLimit )	//flag overload if within OVLimit of max
				m_Overload = TRUE;
			dtmp = pWindowTbl[i];
			pFFTInBuf[j] = dtmp * (double)InBuf[j];	//window the I data
			j++;
			pFFTInBuf[j] = dtmp * (double)InBuf[j];	//window the Q data
			j++;
		}
	}
	else
	{
		for(i=0; i<m_FFTSize; i++)
		{
			if( InBuf[i] > OVLimit )	//flag overload if within OVLimit of max
				m_Overload = TRUE;
			pFFTInBuf[i] = pWindowTbl[i] * (double)InBuf[i];		//window the data
		}
	}
	LeaveCriticalSection(&m_CriticalSection);
	return m_TotalCount;
}

void Cxfft::CalcFFT()
{
	EnterCriticalSection(&m_CriticalSection);
	if(m_Complex)
	{
		//Calculate the complex FFT
		bitrv2(m_FFTSize*2, pWorkArea + 2, pFFTInBuf);
		CpxFFT(m_FFTSize*2, pFFTInBuf, pSinCosTbl);
	}
	else
	{
		//Calculate the real FFT
		bitrv2(m_FFTSize, pWorkArea + 2, pFFTInBuf);
		cftfsub(m_FFTSize, pFFTInBuf, pSinCosTbl);
		rftfsub(m_FFTSize, pFFTInBuf, pWorkArea[1], pSinCosTbl + pWorkArea[0]);
	}
	LeaveCriticalSection(&m_CriticalSection);
}

//////////////////////////////////////////////////////////////////////
//The range is "start" to "stop" and is multiplied by "gain" and copied
//   into "OutBuf[]". Clamped to 0 and -gain.
// The function returns TRUE if the input is overloaded
//////////////////////////////////////////////////////////////////////
BOOL Cxfft::GetFFTData(INT gain, INT start, INT stop, INT* OutBuf )
{
int x;
INT m;
	if(m_Complex)
		m = (m_FFTSize);
	else
		m = (m_FFTSize/2);

INT j;
double sum;
int k;
	if( (stop-start) < (m_PostAverage*2))
		k = (stop-start)/2;
	else
		k = m_PostAverage;
double k1 = (double)(k+1);
int k2;
	k2 = k/2;
int astart,astop;
	if((start - k2)<0)
		astart = k2;
	else
		astart = start;
	if( (stop + k2)>= m)
		astop = m - k2;
	else
		astop = stop;
	if( stop < m)
	{
		EnterCriticalSection(&m_CriticalSection);
		for( INT i=start; i<=stop; i++ )		//copy and scale into OutBuf[]
		{
			if(m_Complex)
			{
				if(m_PostAverage>0)
				{
					if( (i>=astart) && (i<=astop) )
					{
						sum = 0.0;
						for( j=i-k2; j<=i+k2; j++)
						{
							if(m_Invert)
								sum += pFFTAveBuf[(m-j)];
							else
								sum += pFFTAveBuf[j];
						}
						x = (INT)((double)gain*m_dBGainFactor*(sum/k1 - m_dBmaxOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
					else
					{
						if(m_Invert)
							x = (INT)((double)gain*m_dBGainFactor*(pFFTAveBuf[(m-i)] - m_dBmaxOffset));
						else
							x = (INT)((double)gain*m_dBGainFactor*(pFFTAveBuf[i] - m_dBmaxOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
				}
				else
				{
					if(m_Invert)
						x = (INT)((double)gain*m_dBGainFactor*(pFFTAveBuf[(m-i)] - m_dBmaxOffset));
					else
						x = (INT)((double)gain*m_dBGainFactor*(pFFTAveBuf[i] - m_dBmaxOffset));
					if( x < -gain )
						OutBuf[i] = -gain;
					else
						if( x>0 )
							OutBuf[i] = 0;
						else
							OutBuf[i] = x;
				}
			}
			else	//Is a real FFT
			{
				if(m_PostAverage>0)
				{
					if( (i>=astart) && (i<=astop) )
					{
						sum = 0.0;
						for( j=i-k2; j<=i+k2; j++)
						{
							if(m_Invert)
								sum += pFFTAveBuf[(m-j)<<1];
							else
								sum += pFFTAveBuf[j<<1];
						}
						x = (INT)((double)gain*m_dBGainFactor*(sum/k1 - m_dBmaxOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
					else
					{
						sum = 0.0;
						if( i<astart )
						{
							for( j=0; j<k2; j++)
							{
								if(m_Invert)
									sum += pFFTAveBuf[(m-j)<<1];
								else
									sum += pFFTAveBuf[j<<1];
							}
						}
						else
						{
							for( j=astop-k2; j<=astop; j++)
							{
								if(m_Invert)
									sum += pFFTAveBuf[(m-j)<<1];
								else
									sum += pFFTAveBuf[j<<1];
							}
						}
						x = (INT)((double)gain*m_dBGainFactor*(sum/k1 - m_dBmaxOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
				}
				else
				{
					if(m_Invert)
						x = (INT)((double)gain*m_dBGainFactor*(pFFTAveBuf[(m-i)<<1] - m_dBmaxOffset));
					else
						x = (INT)((double)gain*m_dBGainFactor*(pFFTAveBuf[i<<1] - m_dBmaxOffset));
					if( x < -gain )
						OutBuf[i] = -gain;
					else
						if( x>0 )
							OutBuf[i] = 0;
						else
							OutBuf[i] = x;
				}
			}
		}
		LeaveCriticalSection(&m_CriticalSection);
	}
	return m_Overload;
}

//////////////////////////////////////////////////////////////////////
//The range is "start" to "stop" and is multiplied by "gain" and copied
//   into "OutBuf[]". Clamped to 0 and -gain.
// The function returns TRUE if the input is overloaded
//////////////////////////////////////////////////////////////////////
BOOL Cxfft::GetFFTDeltaData(INT gain, INT start, INT stop, INT* OutBuf )
{
	if(!pFFTSecBuf)
		return TRUE;
int x;
INT m;
	if(m_Complex)
		m = (m_FFTSize);
	else
		m = (m_FFTSize/2);
INT j;
double sum;
int k;
	if( (stop-start) < (m_PostAverage*2))
		k = (stop-start)/2;
	else
		k = m_PostAverage;
double k1 = (double)(k+1);
int k2 = k/2;
int astart,astop;
	if((start - k2)<0)
		astart = k2;
	else
		astart = start;
	if( (stop + k2)>= m)
		astop = m - k2;
	else
		astop = stop;
	if( stop < m)
	{
		EnterCriticalSection(&m_CriticalSection);
		for( INT i=start; i<=stop; i++ )		//copy and scale into OutBuf[]
		{
			if(m_Complex)
			{
				if(m_PostAverage>0)
				{
					if( (i>=astart) && (i<=astop) )
					{
						sum = 0.0;
						for( j=i-k2; j<=i+k2; j++)
						{
							if(m_Invert)
								sum = sum + pFFTAveBuf[(m-j)]-pFFTSecBuf[(m-j)];
							else
								sum = sum + pFFTAveBuf[j]-pFFTSecBuf[j];
						}
						x = (INT)((double)gain*m_dBGainFactor*(sum/k1 - m_dBDeltaOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
					else
					{
						if(m_Invert)
							x = (INT)((double)gain*m_dBGainFactor*( (pFFTAveBuf[(m-i)] - pFFTSecBuf[(m-i)] )
															- m_dBDeltaOffset));
						else
							x = (INT)((double)gain*m_dBGainFactor*( (pFFTAveBuf[i] - pFFTSecBuf[i])
															- m_dBDeltaOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
				}
				else
				{
					if(m_Invert)
						x = (INT)((double)gain*m_dBGainFactor*((pFFTAveBuf[(m-i)]- pFFTSecBuf[(m-i)])
														- m_dBDeltaOffset));
					else
						x = (INT)((double)gain*m_dBGainFactor*((pFFTAveBuf[i]- pFFTSecBuf[(i)])
														- m_dBDeltaOffset));
					if( x < -gain )
						OutBuf[i] = -gain;
					else
						if( x>0 )
							OutBuf[i] = 0;
						else
							OutBuf[i] = x;
				}
			}
			else	//Is a real FFT
			{
				if(m_PostAverage>0)
				{
					if( (i>=astart) && (i<=astop) )
					{
						sum = 0.0;
						for( j=i-k2; j<=i+k2; j++)
						{
							if(m_Invert)
								sum = sum + pFFTAveBuf[(m-j)<<1] - pFFTSecBuf[(m-j)<<1];
							else
								sum = sum + pFFTAveBuf[j<<1] - pFFTSecBuf[j<<1];
						}
						x = (INT)((double)gain*m_dBGainFactor*(sum/k1 - m_dBDeltaOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
					else
					{
						if(m_Invert)
							x = (INT)((double)gain*m_dBGainFactor*((pFFTAveBuf[(m-i)<<1]- pFFTSecBuf[(m-i)<<1])
																		- m_dBDeltaOffset));
						else
							x = (INT)((double)gain*m_dBGainFactor*((pFFTAveBuf[i<<1] - pFFTSecBuf[i<<1])
																		- m_dBDeltaOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
				}
				else
				{
					if(m_Invert)
						x = (INT)((double)gain*m_dBGainFactor*((pFFTAveBuf[(m-i)<<1] - pFFTSecBuf[(m-i)<<1])
																	- m_dBDeltaOffset));
					else
						x = (INT)((double)gain*m_dBGainFactor*((pFFTAveBuf[i<<1] - pFFTSecBuf[i<<1])
																	- m_dBDeltaOffset));
					if( x < -gain )
						OutBuf[i] = -gain;
					else
						if( x>0 )
							OutBuf[i] = 0;
						else
							OutBuf[i] = x;
				}
			}
		}
		LeaveCriticalSection(&m_CriticalSection);
	}
	return m_Overload;
}

//////////////////////////////////////////////////////////////////////
//The range is "start" to "stop" and is multiplied by "gain" and copied
//   into "OutBuf[]". Clamped to 0 and -gain.
// The function returns TRUE if the input is overloaded
//////////////////////////////////////////////////////////////////////
BOOL Cxfft::GetFFTPeakData(INT gain, INT start, INT stop, INT* OutBuf )
{
int x;
INT m;
	if(m_Complex)
		m = (m_FFTSize);
	else
		m = (m_FFTSize/2);
INT j;
double sum;
int k;
	if(!pFFTSecBuf)
	{
		for( j=start; j<=stop; j++ )		//copy and scale into OutBuf[]
			OutBuf[j] = 0;
		return m_Overload;		//is not set up for peak calculations
	}

	if( (stop-start) < (m_PostAverage*2))
		k = (stop-start)/2;
	else
		k = m_PostAverage;
double k1 = (double)(k+1);
int k2 = k/2;
int astart,astop;
	if((start - k2)<0)
		astart = k2;
	else
		astart = start;
	if( (stop + k2)>= m)
		astop = m - k2;
	else
		astop = stop;
	if( stop < m)
	{
		EnterCriticalSection(&m_CriticalSection);
		for( INT i=start; i<=stop; i++ )		//copy and scale into OutBuf[]
		{
			if(m_Complex)
			{
				if(m_PostAverage>0)
				{
					if( (i>=astart) && (i<=astop) )
					{
						sum = 0.0;
						for( j=i-k2; j<=i+k2; j++)
						{
							if(m_Invert)
								sum += pFFTSecBuf[(m-j)];
							else
								sum += pFFTSecBuf[j];
						}
						x = (INT)((double)gain*m_dBGainFactor*(sum/k1 - m_dBmaxOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
					else
					{
						if(m_Invert)
							x = (INT)((double)gain*m_dBGainFactor*(pFFTSecBuf[(m-i)] - m_dBmaxOffset));
						else
							x = (INT)((double)gain*m_dBGainFactor*(pFFTSecBuf[i] - m_dBmaxOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
				}
				else
				{
					if(m_Invert)
						x = (INT)((double)gain*m_dBGainFactor*(pFFTSecBuf[(m-i)] - m_dBmaxOffset));
					else
						x = (INT)((double)gain*m_dBGainFactor*(pFFTSecBuf[i] - m_dBmaxOffset));
					if( x < -gain )
						OutBuf[i] = -gain;
					else
						if( x>0 )
							OutBuf[i] = 0;
						else
							OutBuf[i] = x;
				}
			}
			else	//Is a real FFT
			{
				if(m_PostAverage>0)
				{
					if( (i>=astart) && (i<=astop) )
					{
						sum = 0.0;
						for( j=i-k2; j<=i+k2; j++)
						{
							if(m_Invert)
								sum += pFFTSecBuf[(m-j)<<1];
							else
								sum += pFFTSecBuf[j<<1];
						}
						x = (INT)((double)gain*m_dBGainFactor*(sum/k1 - m_dBmaxOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
					else
					{
						if(m_Invert)
							x = (INT)((double)gain*m_dBGainFactor*(pFFTSecBuf[(m-i)<<1] - m_dBmaxOffset));
						else
							x = (INT)((double)gain*m_dBGainFactor*(pFFTSecBuf[i<<1] - m_dBmaxOffset));
						if( x < -gain )
							OutBuf[i] = -gain;
						else
							if( x>0 )
								OutBuf[i] = 0;
							else
								OutBuf[i] = x;
					}
				}
				else
				{
					if(m_Invert)
						x = (INT)((double)gain*m_dBGainFactor*(pFFTSecBuf[(m-i)<<1] - m_dBmaxOffset));
					else
						x = (INT)((double)gain*m_dBGainFactor*(pFFTSecBuf[i<<1] - m_dBmaxOffset));
					if( x < -gain )
						OutBuf[i] = -gain;
					else
						if( x>0 )
							OutBuf[i] = 0;
						else
							OutBuf[i] = x;
				}
			}
		}
		LeaveCriticalSection(&m_CriticalSection);
	}
	return m_Overload;
}

//////////////////////////////////////////////////////////////////////
// The function returns the average value of the fft withing the range
// as well as the peak value
// src == 0 is the main FFT average data
// src == 1 is the secondary(peak or reference) FFT average data
// src == 2 is the delta(main-reference) FFT average data
//////////////////////////////////////////////////////////////////////
double Cxfft::GetFFTave(int src, int start, int stop, double* peak)
{
double sum = 0.0;
double n = 0.0;
double val;
double* pData;
	if(src==0)
		pData = pFFTAveBuf;
	else
		pData = pFFTSecBuf;
	*peak = K_MINDB;
	EnterCriticalSection(&m_CriticalSection);
	if(m_Complex)
	{
		for( int i=start; i<=stop; i++ )
		{
			if(src==2)
				val = 10.0*(pFFTAveBuf[i]-pFFTSecBuf[i]);
			else
				val = 10.0*pData[i];
			sum =  sum + val;
			n += 1.0;
			if( val > *peak )
				*peak = val;
		}
	}
	else
	{
		for( int i=start; i<=stop; i++ )
		{
			if(src==2)
				val = 10.0*(pFFTAveBuf[i<<1]-pFFTSecBuf[i<<1]);
			else
				val = 10.0*pData[i<<1];
			sum =  sum + val;
			n += 1.0;
			if( val > *peak )
				*peak = val;
		}
	}
	LeaveCriticalSection(&m_CriticalSection);
	return sum/n;
}


//////////////////////////////////////////////////////////////////////
// The function returns the raw FFT data from a single bin
// type 0 == ave, 1 = peak, 2 == delta
//////////////////////////////////////////////////////////////////////
double Cxfft::GetSingleBin(INT type, INT bin)
{
double* pBuf;
double val;
INT m;
	if(m_Complex)
		m = (m_FFTSize);
	else
		m = (m_FFTSize/2);
	EnterCriticalSection(&m_CriticalSection);
	switch(type)
	{
		case 0:
			if(m_Complex)
			{
				if(m_Invert)
					val = 10.0*pFFTAveBuf[(m-bin)];
				else
					val = 10.0*pFFTAveBuf[bin];
			}
			else
			{
				if(m_Invert)
					val = 10.0*pFFTAveBuf[(m-bin)<<1];
				else
					val = 10.0*pFFTAveBuf[bin<<1];
			}
			break;
		case 1:
			if(m_Complex)
			{
				if(m_Invert)
					val = 10.0*pFFTSecBuf[(m-bin)];
				else
					val = 10.0*pFFTSecBuf[bin];
			}
			else
			{
				if(m_Invert)
					val = 10.0*pFFTSecBuf[(m-bin)<<1];
				else
					val = 10.0*pFFTSecBuf[bin<<1];
			}
			break;
		case 2:
			if(m_Complex)
			{
				if(m_Invert)
					val = 10.0*(pFFTAveBuf[(m-bin)] - pFFTSecBuf[(m-bin)]);
				else
					val = 10.0*(pFFTAveBuf[bin] - pFFTSecBuf[bin]);
			}
			else
			{
				if(m_Invert)
					val = 10.0*(pFFTAveBuf[(m-bin)<<1] - pFFTSecBuf[(m-bin)<<1]);
				else
					val = 10.0*(pFFTAveBuf[(m-bin)<<1] - pFFTSecBuf[bin<<1]);
			}
			pBuf = pFFTAveBuf;
			break;
	}
	LeaveCriticalSection(&m_CriticalSection);
	return val;
}

//////////////////////////////////////////////////////////////////////
// The function returns TRUE if the input is overloaded
//////////////////////////////////////////////////////////////////////

BOOL Cxfft::GetDoubleFFTData(double* OutBuf )
{
INT i;
INT m;
	EnterCriticalSection(&m_CriticalSection);
	if(m_Complex)
	{
		m = (m_FFTSize) - 1;
		if(m_Invert)
		{
			for( i=0; i<=m; i++ )		//copy and scale inverted spectrum into OutBuf[]
				OutBuf[i] = 10.0*pFFTAveBuf[(m-i)];
		}
		else
		{
			for( i=0; i<=m; i++ )		//copy and scale into OutBuf[]
				OutBuf[i] = 10.0*pFFTAveBuf[i];
		}
	}
	else
	{
		m = (m_FFTSize/2) - 1;
		if(m_Invert)
		{
			for( i=0; i<=m; i++ )		//copy and scale inverted spectrum into OutBuf[]
				OutBuf[i] = 10.0*pFFTAveBuf[(m-i)<<1];
		}
		else
		{
			for( i=0; i<=m; i++ )		//copy and scale into OutBuf[]
				OutBuf[i] = 10.0*pFFTAveBuf[i<<1];
		}
	}
	LeaveCriticalSection(&m_CriticalSection);
	return m_Overload;
}

// Nitty gritty fft routines by Takuya OOURA(Updated to his new version 4-18-02)
// Routine calculates real FFT
void Cxfft::rftfsub(int n, double *a, int nc, double *c)
{
double tmp1,tmp2;
    int j, k, kk, ks, m;
    double wkr, wki, xr, xi, yr, yi;

	m_TotalCount++;
 	if(m_AveCount < m_AveSize)
		m_AveCount++;
    m = n >> 1;
    ks = 2 * nc/m;
    kk = 0;
	for (j = 2; j < m; j += 2 ) 
	{
		k = n - j;
		kk += ks;
		wkr = 0.5 - c[nc - kk];
		wki = c[kk];
		xr = a[j] - a[k];
		xi = a[j + 1] + a[k + 1];
		yr = wkr * xr - wki * xi;
		yi = wkr * xi + wki * xr;
		a[j] -= yr;
		xi = a[j]*a[j];
		a[j+1] -= yi;
		xi += ( a[j+1]*a[j+1]);
		a[k] += yr;
		xr = a[k]*a[k];
		a[k+1] -= yi;
		xr += (a[k+1]*a[k+1]);
		tmp1 = log10(xi+m_K_C) + m_K_B;
		tmp2 = log10(xr+m_K_C) + m_K_B;
		if(m_TotalCount <= m_AveSize)
		{
 			pFFTSumBuf[j] = pFFTSumBuf[j] + tmp1;
 			pFFTSumBuf[k] = pFFTSumBuf[k] + tmp2;
		}
		else
		{
 			pFFTSumBuf[j] = pFFTSumBuf[j] - pFFTAveBuf[j] + tmp1;
 			pFFTSumBuf[k] = pFFTSumBuf[k] - pFFTAveBuf[k] + tmp2;
		}
		pFFTAveBuf[j] = pFFTSumBuf[j]/(double)m_AveCount;
		pFFTAveBuf[k] = pFFTSumBuf[k]/(double)m_AveCount;
		if(m_MaxDelta==MAX_MODE)
		{
			if(	pFFTAveBuf[j] > pFFTSecBuf[j] )
				pFFTSecBuf[j] = pFFTAveBuf[j];
			if(	pFFTAveBuf[k] > pFFTSecBuf[k] )
				pFFTSecBuf[k] = pFFTAveBuf[k];
		}
		else
		{
			if(m_MaxDelta==DELTA_MODE)
			{
				if(m_SaveReference)
				{
					pFFTSecBuf[j] = pFFTAveBuf[j];
					pFFTSecBuf[k] = pFFTAveBuf[k];
				}
			}
		}
	}

	a[0] *= a[0];						//calc DC term
	xr = a[m]*a[m]+a[m+1]*a[m+1];		//calculate N/4(middle) term
	tmp1 = log10(a[0]+m_K_C) + m_K_B;
	tmp2 = log10(xr+m_K_C) + m_K_B;

	if(m_TotalCount <= m_AveSize)
	{
 		pFFTSumBuf[0] = pFFTSumBuf[0] + tmp1;
 		pFFTSumBuf[n/2] = pFFTSumBuf[n/2] + tmp2;
	}
	else
	{
 		pFFTSumBuf[0] = pFFTSumBuf[0] - pFFTAveBuf[0] + tmp1;
 		pFFTSumBuf[n/2] = pFFTSumBuf[n/2] - pFFTAveBuf[n/2] + tmp2;
	}
	pFFTAveBuf[0] = pFFTSumBuf[0]/(double)m_AveCount;
	pFFTAveBuf[n/2] = pFFTSumBuf[n/2]/(double)m_AveCount;
	if(m_MaxDelta==MAX_MODE)
	{
		if(	pFFTAveBuf[0] > pFFTSecBuf[0] )
			pFFTSecBuf[0] = pFFTAveBuf[0];
		if(	pFFTAveBuf[n/2] > pFFTSecBuf[n/2] )
			pFFTSecBuf[n/2] = pFFTAveBuf[n/2];
	}
	else
	{
		if(m_MaxDelta==DELTA_MODE)
		{
			if(m_SaveReference)
			{
				pFFTSecBuf[0] = pFFTAveBuf[0];
				pFFTSecBuf[n/2] = pFFTAveBuf[n/2];
			}
		}
	}
	m_SaveReference = FALSE;
}


void Cxfft::CpxFFT(int n, double *a, double *w)
{
    int j, j1, j2, j3, l;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
	m_TotalCount++;
 	if(m_AveCount < m_AveSize)
		m_AveCount++;
if(m_AveCount==0)
	m_AveCount = 0;
    l = 2;
    if (n > 8) {
        cft1st(n, a, w);
        l = 8;
        while ((l << 2) < n) {
            cftmdl(n, l, a, w);
            l <<= 2;
        }
    }
    if ((l << 2) == n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            a[j2] = x0r - x2r;
            a[j2 + 1] = x0i - x2i;
            a[j1] = x1r - x3i;
            a[j1 + 1] = x1i + x3r;
            a[j3] = x1r + x3i;
            a[j3 + 1] = x1i - x3r;
        }
    } else {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            x0r = a[j] - a[j1];
            x0i = a[j + 1] - a[j1 + 1];
            a[j] += a[j1];
            a[j + 1] += a[j1 + 1];
            a[j1] = x0r;
            a[j1 + 1] = x0i;
        }
    }

	if(m_NCOnull)
	{
		a[0] = (a[4]+a[n-3])/2.;  a[1] = (a[5]+a[n-4])/2.;
		a[2] = a[4];  a[3] = a[5];
		a[n-1] = a[n-3];  a[n-2] = a[n-4];
	}

	for( l=0,j=n/4-1; j>=0; l+=2,j--)
	{
		x0r = (a[l]*a[l]) + (a[l+1]*a[l+1]);
		x0i = log10( x0r + m_K_C) + m_K_B;
		if(m_TotalCount <= m_AveSize)
 			pFFTSumBuf[j] = pFFTSumBuf[j] + x0i;
		else
 			pFFTSumBuf[j] = pFFTSumBuf[j] - pFFTAveBuf[j] + x0i;
		pFFTAveBuf[j] = pFFTSumBuf[j]/(double)m_AveCount;
		if(m_MaxDelta==MAX_MODE)
		{
			if(	pFFTAveBuf[j] > pFFTSecBuf[j] )
				pFFTSecBuf[j] = pFFTAveBuf[j];
		}
		else
		{
			if(m_MaxDelta==DELTA_MODE)
			{
				if(m_SaveReference)
					pFFTSecBuf[j] = pFFTAveBuf[j];
			}
		}
	}
	for( l=n/2,j=(n/2-1); l<n; l+=2,j--)
	{
		x0r = (a[l]*a[l]) + (a[l+1]*a[l+1]);
		x0i = log10( x0r + m_K_C) + m_K_B;
		if(m_TotalCount <= m_AveSize)
 			pFFTSumBuf[j] = pFFTSumBuf[j] + x0i;
		else
 			pFFTSumBuf[j] = pFFTSumBuf[j] - pFFTAveBuf[j] + x0i;
		pFFTAveBuf[j] = pFFTSumBuf[j]/(double)m_AveCount;
		if(m_MaxDelta==MAX_MODE)
		{
			if(	pFFTAveBuf[j] > pFFTSecBuf[j] )
				pFFTSecBuf[j] = pFFTAveBuf[j];
		}
		else
		{
			if(m_MaxDelta==DELTA_MODE)
			{
				if(m_SaveReference)
					pFFTSecBuf[j] = pFFTAveBuf[j];
			}
		}
	}
	m_SaveReference = FALSE;
}

/* -------- initializing routines -------- */
void Cxfft::makewt(int nw, int *ip, double *w)
{
    int j, nwh;
    double delta, x, y;
    
    ip[0] = nw;
    ip[1] = 1;
    if (nw > 2) {
        nwh = nw >> 1;
        delta = atan(1.0) / nwh;
        w[0] = 1;
        w[1] = 0;
        w[nwh] = cos(delta * nwh);
        w[nwh + 1] = w[nwh];
        if (nwh > 2) {
            for (j = 2; j < nwh; j += 2) {
                x = cos(delta * j);
                y = sin(delta * j);
                w[j] = x;
                w[j + 1] = y;
                w[nw - j] = y;
                w[nw - j + 1] = x;
            }
            bitrv2(nw, ip + 2, w);
        }
    }
}

void Cxfft::makect(int nc, int *ip, double *c)
{
    int j, nch;
    double delta;
    
    ip[1] = nc;
    if (nc > 1) {
        nch = nc >> 1;
        delta = atan(1.0) / nch;
        c[0] = cos(delta * nch);
        c[nch] = 0.5 * c[0];
        for (j = 1; j < nch; j++) {
            c[j] = 0.5 * cos(delta * j);
            c[nc - j] = 0.5 * sin(delta * j);
        }
    }
}

/* -------- child routines -------- */
void Cxfft::bitrv2(int n, int *ip, double *a)
{
    int j, j1, k, k1, l, m, m2;
    double xr, xi, yr, yi;
    
    ip[0] = 0;
    l = n;
    m = 1;
    while ((m << 3) < l) {
        l >>= 1;
        for (j = 0; j < m; j++) {
            ip[m + j] = ip[j] + l;
        }
        m <<= 1;
    }
    m2 = 2 * m;
    if ((m << 3) == l) {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 -= m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            j1 = 2 * k + m2 + ip[k];
            k1 = j1 + m2;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
        }
    } else {
        for (k = 1; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
        }
    }
}

void Cxfft::cftfsub(int n, double *a, double *w)
{
    int j, j1, j2, j3, l;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    l = 2;
    if (n > 8) {
        cft1st(n, a, w);
        l = 8;
        while ((l << 2) < n) {
            cftmdl(n, l, a, w);
            l <<= 2;
        }
    }
    if ((l << 2) == n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            a[j2] = x0r - x2r;
            a[j2 + 1] = x0i - x2i;
            a[j1] = x1r - x3i;
            a[j1 + 1] = x1i + x3r;
            a[j3] = x1r + x3i;
            a[j3 + 1] = x1i - x3r;
        }
    } else {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            x0r = a[j] - a[j1];
            x0i = a[j + 1] - a[j1 + 1];
            a[j] += a[j1];
            a[j + 1] += a[j1 + 1];
            a[j1] = x0r;
            a[j1 + 1] = x0i;
        }
    }
}

void Cxfft::cft1st(int n, double *a, double *w)
{
    int j, k1, k2;
    double wk1r, wk1i, wk2r, wk2i, wk3r, wk3i;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    x0r = a[0] + a[2];
    x0i = a[1] + a[3];
    x1r = a[0] - a[2];
    x1i = a[1] - a[3];
    x2r = a[4] + a[6];
    x2i = a[5] + a[7];
    x3r = a[4] - a[6];
    x3i = a[5] - a[7];
    a[0] = x0r + x2r;
    a[1] = x0i + x2i;
    a[4] = x0r - x2r;
    a[5] = x0i - x2i;
    a[2] = x1r - x3i;
    a[3] = x1i + x3r;
    a[6] = x1r + x3i;
    a[7] = x1i - x3r;
    wk1r = w[2];
    x0r = a[8] + a[10];
    x0i = a[9] + a[11];
    x1r = a[8] - a[10];
    x1i = a[9] - a[11];
    x2r = a[12] + a[14];
    x2i = a[13] + a[15];
    x3r = a[12] - a[14];
    x3i = a[13] - a[15];
    a[8] = x0r + x2r;
    a[9] = x0i + x2i;
    a[12] = x2i - x0i;
    a[13] = x0r - x2r;
    x0r = x1r - x3i;
    x0i = x1i + x3r;
    a[10] = wk1r * (x0r - x0i);
    a[11] = wk1r * (x0r + x0i);
    x0r = x3i + x1r;
    x0i = x3r - x1i;
    a[14] = wk1r * (x0i - x0r);
    a[15] = wk1r * (x0i + x0r);
    k1 = 0;
    for (j = 16; j < n; j += 16) {
        k1 += 2;
        k2 = 2 * k1;
        wk2r = w[k1];
        wk2i = w[k1 + 1];
        wk1r = w[k2];
        wk1i = w[k2 + 1];
        wk3r = wk1r - 2 * wk2i * wk1i;
        wk3i = 2 * wk2i * wk1r - wk1i;
        x0r = a[j] + a[j + 2];
        x0i = a[j + 1] + a[j + 3];
        x1r = a[j] - a[j + 2];
        x1i = a[j + 1] - a[j + 3];
        x2r = a[j + 4] + a[j + 6];
        x2i = a[j + 5] + a[j + 7];
        x3r = a[j + 4] - a[j + 6];
        x3i = a[j + 5] - a[j + 7];
        a[j] = x0r + x2r;
        a[j + 1] = x0i + x2i;
        x0r -= x2r;
        x0i -= x2i;
        a[j + 4] = wk2r * x0r - wk2i * x0i;
        a[j + 5] = wk2r * x0i + wk2i * x0r;
        x0r = x1r - x3i;
        x0i = x1i + x3r;
        a[j + 2] = wk1r * x0r - wk1i * x0i;
        a[j + 3] = wk1r * x0i + wk1i * x0r;
        x0r = x1r + x3i;
        x0i = x1i - x3r;
        a[j + 6] = wk3r * x0r - wk3i * x0i;
        a[j + 7] = wk3r * x0i + wk3i * x0r;
        wk1r = w[k2 + 2];
        wk1i = w[k2 + 3];
        wk3r = wk1r - 2 * wk2r * wk1i;
        wk3i = 2 * wk2r * wk1r - wk1i;
        x0r = a[j + 8] + a[j + 10];
        x0i = a[j + 9] + a[j + 11];
        x1r = a[j + 8] - a[j + 10];
        x1i = a[j + 9] - a[j + 11];
        x2r = a[j + 12] + a[j + 14];
        x2i = a[j + 13] + a[j + 15];
        x3r = a[j + 12] - a[j + 14];
        x3i = a[j + 13] - a[j + 15];
        a[j + 8] = x0r + x2r;
        a[j + 9] = x0i + x2i;
        x0r -= x2r;
        x0i -= x2i;
        a[j + 12] = -wk2i * x0r - wk2r * x0i;
        a[j + 13] = -wk2i * x0i + wk2r * x0r;
        x0r = x1r - x3i;
        x0i = x1i + x3r;
        a[j + 10] = wk1r * x0r - wk1i * x0i;
        a[j + 11] = wk1r * x0i + wk1i * x0r;
        x0r = x1r + x3i;
        x0i = x1i - x3r;
        a[j + 14] = wk3r * x0r - wk3i * x0i;
        a[j + 15] = wk3r * x0i + wk3i * x0r;
    }
}

void Cxfft::cftmdl(int n, int l, double *a, double *w)
{
    int j, j1, j2, j3, k, k1, k2, m, m2;
    double wk1r, wk1i, wk2r, wk2i, wk3r, wk3i;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    m = l << 2;
    for (j = 0; j < l; j += 2) {
        j1 = j + l;
        j2 = j1 + l;
        j3 = j2 + l;
        x0r = a[j] + a[j1];
        x0i = a[j + 1] + a[j1 + 1];
        x1r = a[j] - a[j1];
        x1i = a[j + 1] - a[j1 + 1];
        x2r = a[j2] + a[j3];
        x2i = a[j2 + 1] + a[j3 + 1];
        x3r = a[j2] - a[j3];
        x3i = a[j2 + 1] - a[j3 + 1];
        a[j] = x0r + x2r;
        a[j + 1] = x0i + x2i;
        a[j2] = x0r - x2r;
        a[j2 + 1] = x0i - x2i;
        a[j1] = x1r - x3i;
        a[j1 + 1] = x1i + x3r;
        a[j3] = x1r + x3i;
        a[j3 + 1] = x1i - x3r;
    }
    wk1r = w[2];
    for (j = m; j < l + m; j += 2) {
        j1 = j + l;
        j2 = j1 + l;
        j3 = j2 + l;
        x0r = a[j] + a[j1];
        x0i = a[j + 1] + a[j1 + 1];
        x1r = a[j] - a[j1];
        x1i = a[j + 1] - a[j1 + 1];
        x2r = a[j2] + a[j3];
        x2i = a[j2 + 1] + a[j3 + 1];
        x3r = a[j2] - a[j3];
        x3i = a[j2 + 1] - a[j3 + 1];
        a[j] = x0r + x2r;
        a[j + 1] = x0i + x2i;
        a[j2] = x2i - x0i;
        a[j2 + 1] = x0r - x2r;
        x0r = x1r - x3i;
        x0i = x1i + x3r;
        a[j1] = wk1r * (x0r - x0i);
        a[j1 + 1] = wk1r * (x0r + x0i);
        x0r = x3i + x1r;
        x0i = x3r - x1i;
        a[j3] = wk1r * (x0i - x0r);
        a[j3 + 1] = wk1r * (x0i + x0r);
    }
    k1 = 0;
    m2 = 2 * m;
    for (k = m2; k < n; k += m2) {
        k1 += 2;
        k2 = 2 * k1;
        wk2r = w[k1];
        wk2i = w[k1 + 1];
        wk1r = w[k2];
        wk1i = w[k2 + 1];
        wk3r = wk1r - 2 * wk2i * wk1i;
        wk3i = 2 * wk2i * wk1r - wk1i;
        for (j = k; j < l + k; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            x0r -= x2r;
            x0i -= x2i;
            a[j2] = wk2r * x0r - wk2i * x0i;
            a[j2 + 1] = wk2r * x0i + wk2i * x0r;
            x0r = x1r - x3i;
            x0i = x1i + x3r;
            a[j1] = wk1r * x0r - wk1i * x0i;
            a[j1 + 1] = wk1r * x0i + wk1i * x0r;
            x0r = x1r + x3i;
            x0i = x1i - x3r;
            a[j3] = wk3r * x0r - wk3i * x0i;
            a[j3 + 1] = wk3r * x0i + wk3i * x0r;
        }
        wk1r = w[k2 + 2];
        wk1i = w[k2 + 3];
        wk3r = wk1r - 2 * wk2r * wk1i;
        wk3i = 2 * wk2r * wk1r - wk1i;
        for (j = k + m; j < l + (k + m); j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            x0r -= x2r;
            x0i -= x2i;
            a[j2] = -wk2i * x0r - wk2r * x0i;
            a[j2 + 1] = -wk2i * x0i + wk2r * x0r;
            x0r = x1r - x3i;
            x0i = x1i + x3r;
            a[j1] = wk1r * x0r - wk1i * x0i;
            a[j1 + 1] = wk1r * x0i + wk1i * x0r;
            x0r = x1r + x3i;
            x0i = x1i - x3r;
            a[j3] = wk3r * x0r - wk3i * x0i;
            a[j3 + 1] = wk3r * x0i + wk3i * x0r;
        }
    }
}

int _matherr( struct _exception *except )
{
testflag = TRUE;
    /* Handle _DOMAIN errors for log or log10. */
    if( except->type == _DOMAIN )
    {
TRACE("FP Error\n");
    }
    else
    {
        return 0;    /* Else use the default actions */
    }
    return 0;    /* Else use the default actions */
}




